/*
 *  C++ Performance API wrapper.
 *
 *  Copyright 2012 Aurelian Melinte. 
 *  Released under GPL 3.0 or later. 
 *
 *  You need a C++0x compiler. 
 * 
 */


#include <iostream>
#include <stdexcept>
#include <string>
#include <algorithm>
#include <vector>
#include <list>
#include <cassert>
#include <cstdatomic>
#include <mutex>

#include <pthread.h>

#include "papi.h"

#ifndef _PAPI_HPP_
#define _PAPI_HPP_

namespace papi {

/*
 *
 */

class event
{
public:

    event(int num, const char* info) 
        : _init(false)
        , _num(num)
        , _info(info)
    {}
    
   
    int num() const { return _num; }
    const char * name() const { init(); return _name; }
    const char * info() const 
    { 
        init(); 
        return _info ? _info : _einfo.short_descr;
    }

private: 

    void init() const 
    {
        if (!_init) {
            PAPI_event_code_to_name(_num, _name);
            PAPI_get_event_info(_num, &_einfo);
            
            _init = true;
        }
    }
    
private: 

    mutable bool _init;

    int _num;
    const char* _info;
    mutable char _name[PAPI_MAX_STR_LEN+1];
    mutable PAPI_event_info_t _einfo;
};

/*
 * What will PAPI be watching over
 */

typedef std::vector<event> events_t;
static  const events_t events = {  
      {PAPI_TOT_INS, "Total instructions"} 
    , {PAPI_TOT_CYC, "Total cpu cycles"}
    , {PAPI_L1_DCM,  "L1 load  misses"}
//  , {PAPI_L1_STM,  "L1 store  missess"}
    , {PAPI_L2_DCM,  "L2 load  misses"}
//  , {PAPI_L2_STM,  "L2 store  missess"}
    , {PAPI_BR_MSP,  "Branch mispredictions"}
};
static const int num_events = events.size();



/*
 *
 */

class papi_error : public std::runtime_error
{
public: 

    explicit papi_error(std::string const& msg, int err) 
        : std::runtime_error(msg)
        , _msg(msg)
        , _err(err)
    {
        _msg += ": ";
        _msg += PAPI_strerror(_err);
    }
    
    papi_error(papi_error&& other) throw()
        : std::runtime_error(std::move(other._msg.c_str()))
        , _msg(std::move(other._msg))
        , _err(other._err)
    {
        ///swap(other);
    }
    
    papi_error& operator=(papi_error other) throw()
    {
        swap(other);
        return *this;
    }
    
    papi_error& operator=(papi_error&& other) throw()
    {
        ///swap(other);
        std::runtime_error::operator=(std::move(other));
        _msg = std::move(other._msg); 
        _err = other._err; 
        return *this;
    }
    
    virtual ~papi_error() throw() {}
    
    const char *what() const throw()
    {
        return _msg.c_str();
    }

private: 

    //FIXME: namespace level swap
    void swap(papi_error& other) throw()
    {
        std::swap(_msg, other._msg);
        _err = other._err;
    }

private: 

    std::string _msg;
    int _err;
};



/*
 *
 */

class pmutex_guard
{
public:

    pmutex_guard(pthread_mutex_t* mutex)
        : _mutex(mutex)
    {
        int ret = pthread_mutex_lock(_mutex);
        if (-1 == ret)
            throw std::runtime_error("pthread_mutex_lock");
    }
    
    ~pmutex_guard()
    {
        int ret = pthread_mutex_unlock(_mutex);
    }


private:

    pmutex_guard();
    pmutex_guard(pmutex_guard const&);
    pmutex_guard& operator=(pmutex_guard const&);

private:

    pthread_mutex_t * _mutex;
};


/*
 * Policies
 */

struct counting_per_thread
{
    struct process 
    {
        static void apply(int event_set)
        {
#if 0 
            int retval = PAPI_OK;
            if ( (retval = PAPI_start(event_set)) != PAPI_OK) {
                throw papi_error("PAPI_start", retval);
            }
#endif        
        }
    };
    struct thread 
    {
        static void apply(int event_set)
        {
            int retval = PAPI_OK;
            if ( (retval = PAPI_start(event_set)) != PAPI_OK) {
                throw papi_error("PAPI_start", retval);
            }
        }
    };

protected:
    ~counting_per_thread() {}
};


struct multiplex_none
{
    struct process 
    {
        static void apply(int event_set)
        {
#if 0
            int retval = PAPI_OK;
            if ( (retval = PAPI_multiplex_init()) != PAPI_OK) {
                throw papi_error("PAPI_multiplex_init", retval);
            }
#endif           
        }
    };
    struct thread 
    {
        static void apply(int event_set)
        {
#if 0 
            int retval = PAPI_OK;
            if ( (retval = PAPI_set_multiplex(event_set)) != PAPI_OK) {
                throw papi_error("PAPI_set_multiplex", retval);
            }
#endif
        }
    };

protected:
    ~multiplex_none() {}
};


/*
 *
 */

class library
{
public:

    template < class Counting 
             , class Multiplex 
             > 
    static bool init()
    {
        if (!_lib_init.load()) {
            std::unique_lock<std::mutex> lock(_lib_mutex);
            
            if (_lib_init.load())
                return true;
        
            int num_hwcntrs = 0;
            int retval = 0;

            if ((retval = PAPI_library_init(PAPI_VER_CURRENT)) != PAPI_VER_CURRENT ) {
                throw papi_error("PAPI_library_init", retval);
            }
        
            Multiplex::process::apply(_event_set);

            if ( (retval = PAPI_thread_init( (unsigned long (*)(void))(pthread_self) )) != PAPI_OK) {
                throw papi_error("PAPI_thread_init", retval);
            }
           
            /*
             * PAPI_num_counters returns the number of hardware counters the platform
             * has or a negative number if there is an error 
             */
            if ((num_hwcntrs = PAPI_num_counters()) < PAPI_OK) {
                throw papi_error("PAPI_num_counters", retval);
            }

            const PAPI_hw_info_t *hwinfo = PAPI_get_hardware_info();
            if (hwinfo == NULL) {
                throw papi_error("PAPI_get_hardware_info", 0);
            }
            
            std::cout <<  "There are " << num_hwcntrs 
                      << " counters for " << hwinfo->vendor_string 
                      << ":" << hwinfo->model_string 
                      << std::endl;
            std::cout <<  "Cache: " << hwinfo->mem_hierarchy.levels << " levels\n";
            for (int i = 0; i < hwinfo->mem_hierarchy.levels; ++i) {
                const PAPI_mh_level_t& level = hwinfo->mem_hierarchy.level[i];
                std::cout << " " << i+1 << ".1:  size=" << level.cache[0].size 
                          << " " << level.cache[0].num_lines << " lines * " 
                          << level.cache[0].line_size << " assoc " 
                          << level.cache[0].associativity << "\n"
                          ;
                std::cout << " " << i+1 << ".1:  size=" << level.cache[1].size 
                          << " " << level.cache[1].num_lines << " lines * " 
                          << level.cache[1].line_size << " assoc " 
                          << level.cache[1].associativity << "\n"
                          ;
            }
            std::cout << std::endl;

            if ( (retval = PAPI_create_eventset(&_event_set)) != PAPI_OK) {
                throw papi_error("PAPI_create_eventset", retval);
            }

            for (std::vector<event>::const_iterator it = events.begin();
                 it != events.end();
                 ++it) {
                if ((retval=PAPI_add_event(_event_set, it->num())) != PAPI_OK) {
                    throw papi_error("PAPI_add_event", retval);
                }
            }
            
            Counting::process::apply(_event_set);

            _lib_init.store(true);
        }
    
        return _lib_init.load();
    }
    
    static int event_set()
    {
        return _event_set;
    }
    
private:

    static std::atomic<bool> _lib_init;// = false;
    static std::mutex _lib_mutex;
    static int  _event_set; // = PAPI_NULL;
    
    //FIXME:
    //  PAPI_cleanup_eventset
    //  PAPI_shutdown
};


/*
 *
 */

class thread
{
public:

    template < class Counting 
             , class Multiplex 
             > 
    static void init()
    {
        if (static_cast<unsigned long>(-1) == _tid) {
            library::init<Counting, Multiplex >();
            
            int retval = 0;

            /*
             * Cf. PAPI Users Manual, unbounded threads use per process counters.
             * Still...
             */
            if ( (retval = PAPI_register_thread()) != PAPI_OK) {
                throw papi_error("PAPI_register_thread", retval);
            }

            if ( (retval = PAPI_create_eventset(&thread::_event_set)) != PAPI_OK) {
                throw papi_error("PAPI_create_eventset", retval);
            }

            for (std::vector<event>::const_iterator it = events.begin();
                 it != events.end();
                 ++it) {
                if ((retval=PAPI_add_event(_event_set, it->num())) != PAPI_OK) {
                    throw papi_error("PAPI_add_event", retval);
                }
            }
        
            Multiplex::thread::apply(_event_set);

            Counting::thread::apply(_event_set);

            _tid = PAPI_thread_id();
        }
    }
    
    static int event_set()
    {
        return _event_set;
    }

private:

    thread();
    thread(thread const&);
    thread& operator=(thread const&);
    
    //FIXME:
    //  PAPI_stop
    //  PAPI_cleanup_eventset
    //  PAPI_unregister_thread

private:

    static __thread unsigned long _tid; // = (unsigned long)-1;
    static __thread int  _event_set; // = PAPI_NULL;
};


/*
 * Operations on counters
 */

struct nop
{
    template < typename Operand > 
    static void apply(Operand const& c)
    {
    }

protected:
    ~nop() {}
};

struct stdout_print
{
    template < typename Operand > 
    static void apply(Operand const& c)
    {
        std::cout << c;
    }

protected:
    ~stdout_print() {}
};



/*
 *
 */

class counters_const_iterator;
class counter
{
public:

    typedef long long             value_t;
    typedef std::vector<value_t>  values_t;

    explicit counter(const events_t::const_iterator& ev, 
                     const values_t::const_iterator& b, 
                     const values_t::const_iterator& e)
        : _event(ev)
        , _begin(b)
        , _end(e)
    {}
    
    int num() const  { return _event->num(); }
    const char * name() const { return _event->name(); }
    const char * info() const { return _event->info(); }
    value_t begin() const { return *_begin; }
    value_t end() const { return *_end; }
    
    void print(std::ostream& os) const
    {
        os << name() << " (" <<  info() << "): " << end() - begin() 
           << " (" << end() <<"-"<< begin() << ")";
    }
    
    friend std::ostream& operator<< (std::ostream&, counter const&);
    friend class counters_const_iterator;
    
private:

    events_t::const_iterator  _event;
    values_t::const_iterator  _begin;
    values_t::const_iterator  _end;
};


inline std::ostream& operator<< (std::ostream& os, counter const& c)
{
    c.print(os);
}


class counters_const_iterator 
    : public std::iterator< std::bidirectional_iterator_tag
                          , ptrdiff_t 
                          >
{
    counter _counter;

public:
    counters_const_iterator( events_t::const_iterator const& ev
                           , counter::values_t::const_iterator const& b
                           , counter::values_t::const_iterator const& e
                           )
        : _counter(ev, b, e) 
        {}

    bool operator==(const counters_const_iterator& x) const 
    {
        return    _counter._event == x._counter._event 
               && _counter._begin == x._counter._begin 
               && _counter._end == x._counter._end;
    }

    bool operator!=(const counters_const_iterator& x) const 
    {
        return !(*this == x);
    }

    const counter& operator*() const 
    {
        return _counter;
    }
    const counter* operator->() const 
    {
        return &_counter;
    }

    counters_const_iterator& operator++() 
    {
        ++_counter._event;  
        ++_counter._begin;  
        ++_counter._end;
        return *this;
    }
    counters_const_iterator operator++(int) 
    {
        counters_const_iterator tmp = *this;
        ++_counter._event;  
        ++_counter._begin;  
        ++_counter._end;
        return tmp;
    }

    counters_const_iterator& operator--() 
    {
        --_counter._event;  
        --_counter._begin;  
        --_counter._end;
        return *this;
    }
    counters_const_iterator operator--(int) 
    {
        counters_const_iterator tmp = *this;
        --_counter._event;  
        --_counter._begin;  
        --_counter._end;
        return tmp;
    }
}; //iterator


template < typename Op
         , typename Counting = counting_per_thread
         , typename Multiplex = multiplex_none
         >
class counters 
    : public Op
    , public Counting
    , public Multiplex
{

public: 

    counters(std::string const& tag) : _tag(tag)
    {
        thread::init<Counting, Multiplex >();
        
        _start_vals.resize(num_events);
        _end_vals.resize(num_events);
        
        // Note: PAPI_read_counters() resets the counters.
        int retval = 0;    
        if ( (retval=PAPI_read(thread::event_set(), &_start_vals[0])) != PAPI_OK) {
            throw papi_error("PAPI_read", retval);
        }
    }
    
    ~counters()
    {
        int retval = 0;    
        if ( (retval=PAPI_read(thread::event_set(), &_end_vals[0])) != PAPI_OK) {
            throw papi_error("PAPI_read", retval);
        }
    
        Op::apply(*this);
    }
    
    
    typedef counters_const_iterator const_iterator;
    const_iterator begin() const { return const_iterator(events.begin(), 
                                                         _start_vals.begin(), 
                                                         _end_vals.begin()); 
                                 }
    const_iterator end() const   { return const_iterator(events.end(), 
                                                         _start_vals.end(), 
                                                         _end_vals.end()); 
                                 }

private:

    counters();
    counters(const counters<Op, Counting, Multiplex>&);
    counters& operator=(const counters<Op, Counting, Multiplex>&);
    
    // Only on the stack for now. 
    static void* operator new (size_t size);
    static void* operator new (size_t size, void* mem);
    
    template < class O, 
               class C, 
               class M 
             > 
    friend std::ostream& operator<< (std::ostream&, counters<O, C, M> const&);

private:

   counter::values_t  _start_vals;
   counter::values_t  _end_vals;
   std::string _tag;
};


struct counter_printer {
    counter_printer(std::ostream& os) : _os(os) {}
        
    void operator()(counter const& c) 
    {
        _os << "  ";
        c.print(_os);
        _os << "\n";
    }
        
private:
    std::ostream& _os;
};
    
template < class Op
         , class Counting
         , class Multiplex 
         >
std::ostream& operator<< (std::ostream& os, counters<Op, Counting, Multiplex > const& cs)
{
    os << "\nDelta " << cs._tag << ":\n";
    std::for_each(cs.begin(),
                  cs.end(),
                  counter_printer(os) //[&] (counter const& c) { c.print(os);} 
                  );
    os << std::endl;
}


} //namespace


#endif //_PAPI_HPP_
